﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.DesktopEdition;
using System.Collections.ObjectModel;
using System.Diagnostics;


namespace ThinkGeo_ZeroHeadingError
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {

            winformsMap1.MapUnit = GeographyUnit.Meter;

            ManagedProj4Projection projSphericalMercator = new ManagedProj4Projection();
            projSphericalMercator.InternalProjectionParametersString = ManagedProj4Projection.GetEpsgParametersString(4326);
            projSphericalMercator.ExternalProjectionParametersString = ManagedProj4Projection.GetSphericalMercatorParametersString();

            //Sets the zoom levels and styles for countries layer.
            ShapeFileFeatureLayer LineLayer = new ShapeFileFeatureLayer(@"../../data/LineZeroHeading.shp");
            LineLayer.ZoomLevelSet.ZoomLevel01.DefaultLineStyle = LineStyles.River1;
            LineLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            LineLayer.FeatureSource.Open();
            LineLayer.FeatureSource.Projection = projSphericalMercator;
            LineLayer.FeatureSource.Close();

            ShapeFileFeatureLayer PointsLayer = new ShapeFileFeatureLayer(@"../../data/PointsZeroHeading.shp");
            PointsLayer.ZoomLevelSet.ZoomLevel01.DefaultPointStyle = PointStyles.CreateSimpleCircleStyle(GeoColor.StandardColors.Red, 8);
            PointsLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            PointsLayer.FeatureSource.Open();
            PointsLayer.FeatureSource.Projection = projSphericalMercator;
            PointsLayer.FeatureSource.Close();


            LayerOverlay cultureOverlay = new LayerOverlay();
            cultureOverlay.Layers.Add("LinesLayer", LineLayer);
            cultureOverlay.Layers["LinesLayer"].Name = "LinesLayer";
            cultureOverlay.Layers.Add("PointsLayer", PointsLayer);
            cultureOverlay.Layers["PointsLayer"].Name = "PointsLayer";
            winformsMap1.Overlays.Add("CultureOverlay", cultureOverlay);
            winformsMap1.Overlays["CultureOverlay"].Name = "CultureOverlay";

            Layer selectedLayer = ((LayerOverlay)winformsMap1.Overlays["CultureOverlay"]).Layers["LinesLayer"];
            TrackToExtent(winformsMap1, selectedLayer);

            SetupGraticule();

        }

        public static void TrackToExtent(WinformsMap winformsMap, Layer layer)
        {
            if (layer.HasBoundingBox)
            {
                layer.Open();
                RectangleShape extent = layer.GetBoundingBox();
                layer.Close();

                winformsMap.CurrentExtent = ExtentHelper.GetDrawingExtent(extent, winformsMap.Width, winformsMap.Height);

                winformsMap.Refresh();
            }
        }

        public void SetupGraticule()
        {
            GraticuleAdornmentLayer graticuleAdornmentLayer = new GraticuleAdornmentLayer();
            graticuleAdornmentLayer.GraticuleTextFont = new GeoFont("Arial", 8, DrawingFontStyles.Bold);

            winformsMap1.AdornmentOverlay.Layers.Add("graticule", graticuleAdornmentLayer);
                
            Proj4Projection mgdproj4Projection = new Proj4Projection();
            mgdproj4Projection.InternalProjectionParametersString = ManagedProj4Projection.GetWgs84ParametersString();
            mgdproj4Projection.ExternalProjectionParametersString = ManagedProj4Projection.GetSphericalMercatorParametersString();
            mgdproj4Projection.Open();
            graticuleAdornmentLayer.Projection = mgdproj4Projection;
            mgdproj4Projection.Close();

            winformsMap1.Refresh();

        }

        private void btnReturnAllPointsNearALine_Click(object sender, EventArgs e)
        {

            string utmZone = "24";
            string utmHemi = "south";

            Proj4Projection proj4Projection = new Proj4Projection();
            //proj4Projection.InternalProjectionParametersString = Proj4Projection.GetEpsgParametersString(4326); //Geodetic (Longitude/Latitude)
            proj4Projection.ExternalProjectionParametersString = string.Format("+proj=utm +zone={0} +{1} +ellps=WGS84 +datum=WGS84 +units=m +no_defs", utmZone, utmHemi);//UTM
            proj4Projection.InternalProjectionParametersString = ManagedProj4Projection.GetSphericalMercatorParametersString();

            MessageBox.Show(FindAllRxInALine(winformsMap1, proj4Projection, Convert.ToDouble(textSearchDistance.Text))
                , "Items not found", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        }

        public static Feature ConvertSMToUTMCoordinate(Feature feature, Proj4Projection utmProjection)
        {
            // Convert the feature to Lat / Long from Spherical Mercator
            Proj4Projection mgdproj4Projection = new Proj4Projection();
            mgdproj4Projection.InternalProjectionParametersString = utmProjection.ExternalProjectionParametersString;
            mgdproj4Projection.ExternalProjectionParametersString = ManagedProj4Projection.GetSphericalMercatorParametersString();
            mgdproj4Projection.Open();
            Feature geoFeature = mgdproj4Projection.ConvertToInternalProjection(feature);
            mgdproj4Projection.Close();

            return geoFeature;

        }

        private static string FindAllRxInALine(WinformsMap winformsMap, Proj4Projection proj4Projection,
            double maximumSearchDistance)
        {


            // Get a list of all the featureNames
            List<string> featuresToSearch = new List<string>();

            // Create an in memory layer of all the Points.  Be sure to put them in UTM Grid
            InMemoryFeatureLayer pointsIMFL = new InMemoryFeatureLayer();
            pointsIMFL.Open();
            pointsIMFL.Columns.Add(new FeatureSourceColumn("Name"));
            pointsIMFL.Close();

            FeatureLayer rxFeatureLayer = winformsMap.FindFeatureLayer("PointsLayer");
            rxFeatureLayer.Open();
            foreach (Feature feature in rxFeatureLayer.FeatureSource.GetAllFeatures(ReturningColumnsType.AllColumns))
            {
                string featureName = feature.ColumnValues["Name"];
                Debug.Print(feature.GetWellKnownText());
                featuresToSearch.Add(featureName);
                Feature convertedFeature = ConvertSMToUTMCoordinate(feature, proj4Projection);
                Debug.Print(convertedFeature.GetWellKnownText());
                pointsIMFL.InternalFeatures.Add(featureName, convertedFeature);
            }
            rxFeatureLayer.Close();

            int originalFeatureCount = pointsIMFL.InternalFeatures.Count;

            // Create an in memory layer of all the Lines.  Be sure to put them in UTM Grid
            InMemoryFeatureLayer linesIMFL = new InMemoryFeatureLayer();
            linesIMFL.Open();
            linesIMFL.Columns.Add(new FeatureSourceColumn("Name"));
            linesIMFL.Close();

            FeatureLayer txFeatureLayer = winformsMap.FindFeatureLayer("LinesLayer");

            txFeatureLayer.Open();
            foreach (Feature feature in txFeatureLayer.FeatureSource.GetAllFeatures(ReturningColumnsType.AllColumns))
            {
                string featureName = feature.ColumnValues["Name"];
                Debug.Print(feature.GetWellKnownText());
                Feature convertedFeature = ConvertSMToUTMCoordinate(feature, proj4Projection);
                Debug.Print(convertedFeature.GetWellKnownText());
                linesIMFL.InternalFeatures.Add(featureName, convertedFeature);
            }
            txFeatureLayer.Close();


            Collection<Collection<Feature>> allRxLines = new Collection<Collection<Feature>>();

            // First, use the existing Tx lines to check for groups of receivers that fall in or near them.
            // Just make the search distance smaller than 1/2 the crossline spacing interval
            //double maximumSearchDistance = (crosslineSpacing / 2) - 50;

            foreach (Feature feature in linesIMFL.InternalFeatures)
            {

                // Make a line shape from the tx Feature
                MultilineShape multiLine = new MultilineShape(feature.GetWellKnownText());
                LineShape txLine = multiLine.Lines[0];

                // Query all the receivers for any that are less than 1/2 the crossline distance
                // Note that in situations where lines are right in the middle of receivers, we want to 
                // run the below Rx line based code.
                pointsIMFL.Open();
                Collection<Feature> spatialQueryResults = pointsIMFL.QueryTools.GetFeaturesWithinDistanceOf(txLine, GeographyUnit.Meter,
                    DistanceUnit.Meter, maximumSearchDistance, ReturningColumnsType.AllColumns);
                pointsIMFL.Close();

                if (spatialQueryResults.Count > 0)
                {
                    Collection<Feature> rxOnALine = new Collection<Feature>();
                    // But, we also need to remove these items from the features to search list
                    List<string> namesToRemove = new List<string>();
                    foreach (Feature queryFeature in spatialQueryResults)
                    {
                        string queriedFeatureName = queryFeature.ColumnValues["Name"];
                        foreach (string name in featuresToSearch)
                        {
                            if (name == queriedFeatureName)
                            {
                                rxOnALine.Add(queryFeature);
                                namesToRemove.Add(name);
                                break;
                            }
                        }
                    }

                    // Now remove the names from the feature layer
                    foreach (string nameToRemove in namesToRemove)
                    {
                        pointsIMFL.InternalFeatures.Remove(nameToRemove);
                        featuresToSearch.Remove(nameToRemove);
                    }

                    // Add all these Rx to a line collection item
                    allRxLines.Add(rxOnALine);
                }

            }

            // If there are any Rx left in the list, we need to add them to their own receiver line by trying
            // to fit them on their own projected line.  This may be from a cross or just spot receivers added.

            string remainingNames = string.Empty;
            foreach (string remainingFeature in featuresToSearch)
            {
                remainingNames = remainingNames + remainingFeature + ", ";
            }

            if (remainingNames.Length > 0)
            {
                remainingNames = remainingNames.Substring(0, remainingNames.Length - 2);

                return string.Format("Out of {0} points, {1} were not found by the query function"
                    + "\n\nFeature Names are {2}", originalFeatureCount, featuresToSearch.Count, remainingNames);
            }
            else
            {
                return string.Format("No points were missing from the query");
            }
        }



    }
}
